from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, List
from dataclasses import dataclass, field

from igi_gc_reader.reader.excel_address import ExcelAddress, col_index_to_letter, col_letter_to_index
from igi_gc_reader.reader.excel_reader import IExcelReader


class GcFileClass(Enum):
    One = 1
    #Two = 2
    Unclassified = 99


@dataclass
class AnalysisSearchParams:
    start_col: str = 'AJ'   # lowest seen: AM
    end_col: str = 'BG'     # highest seen: BE
    start_row: int = 1      # rowest seen: 1
    end_row: int = 3        # highest seen: 3


# for some file formats the main data headers start in the first couple of rows
# so just detecting a value e.g. in BE1 isn't enough to classify - we need to know
# whether it is an analysis. examples of analysis I have seen are: "WHOLE OIL GC", 
# "AROMATIC GCMS", "SATURATE GCMS", "Saturate Biomarkers", "AROMATIC BIOMARKERS",
# "Hight Temp GC" and "Extract GC". + Saturate GCMSMS
treat_as_analysis_if_contains = ["gc", "saturate", "arom", "biom", "whole", 'msms']


def is_an_analysis(val: Any) -> bool:
    val = str(val).lower()
    return any(snippet in val for snippet in treat_as_analysis_if_contains)


@dataclass
class IClassification(ABC):
    file_class: GcFileClass
    analysis: str = field(default_factory=lambda: "")
    analysis_addr: str = field(default_factory=lambda: "")

    @abstractmethod
    def is_instance(self, xl_reader: IExcelReader) -> bool:
        pass


# TODO - (MAYBE) *** Future note for classifier ***
#        distinguishing characteristic of this file class seems to be 🤞 the repeating metadata
#        blocks at the top of each page - looking for any of 4 specific headers usually seen in 
#        left metadata column. This could be a useful test when implementing a better classifier!


class ClassificationOne(IClassification):
    file_class: GcFileClass = field(default=GcFileClass.One)

    def is_instance(self, xl_reader: IExcelReader) -> bool:
        start_col_idx = col_letter_to_index(AnalysisSearchParams.start_col)
        end_col_idx = col_letter_to_index(AnalysisSearchParams.end_col)
        for row in range(AnalysisSearchParams.start_row, AnalysisSearchParams.end_row + 1):
            for col_idx in range(start_col_idx, end_col_idx + 1):
                col = col_index_to_letter(col_idx)
                addr = ExcelAddress(f'{col}{row}')
                val = xl_reader.read_cell(addr)
                if is_an_analysis(val):
                    self.analysis = val
                    self.analysis_addr = addr.addr
                    return True
        return False


class NoClassification(IClassification):
    file_class: GcFileClass = field(default=GcFileClass.Unclassified)

    def is_instance(self, _: IExcelReader) -> bool:
        raise NotImplementedError("Rather than calling is_instance - use this class if you "
            "have tried all registered classifiers and found no matches")


registered_classifications: List[IClassification] = [
    ClassificationOne(GcFileClass.One),
    ]


def classify(xl_reader: IExcelReader) -> IClassification:
    for classification in registered_classifications:
        if classification.is_instance(xl_reader):
            return classification
    return NoClassification(file_class=GcFileClass.Unclassified)



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, ClassVar, DefaultDict, Dict, Iterator, List, Tuple, Set
from abc import ABC, abstractmethod
from enum import Enum
from functools import reduce

from igi_gc_reader.reader.excel_address import ExcelAddress, col_index_to_letter, col_letter_to_index
from igi_gc_reader.reader.excel_reader import IExcelReader, has_value
from igi_gc_reader.reader.classification import GcFileClass
from igi_gc_reader.utils import gc_logger


@dataclass
class GcPageData:
    """Can represent a single page or merged pages."""
    page_no: str
    headers: List[str] = field(default_factory=list)
    data_rows: List[List[Any]] = field(default_factory=list)

    def has_data(self) -> bool:
        return any(self.data_rows)


@dataclass  # type: ignore
class IGcPageReader(ABC):
    _xl_reader: IExcelReader

    @abstractmethod
    def get_pages(self) -> Iterator[GcPageData]:
        pass


class PagePhase(Enum):
    Seeking = 1
    InPageMetadata = 2
    BetweenMetadataAndHeaders = 3
    InPageHeaders = 4
    InPageData = 5


# TODO - (MAYBE) Consider changing to a pattern where 
#        we have an IPageState abstract class and a subclass for each state:

# IPageState(xl_reader, page_data):
#   def transition(row) -> phase  # switch to alt state if transition criteria is detected
#   def process(row)

# then just go through states for each row and call transition (and process if no transition)


# See readme.md at top level dir for overview of page handling logic
# --- ---------
@dataclass
class GcPageClassOneReader(IGcPageReader):
    _xl_reader: IExcelReader
    header_cols: List[str] = field(default_factory=list)  # assume data uses same cols for now
    expected_metadata_headers: ClassVar[Set[str]] = {'Company:', 'Well Name:', 'Depth:', 
                                                      'Sampling Point:'}
    page_start_col: ClassVar[str] = 'A'
    post_data_headers: ClassVar[Set[str]] = {  # use as an indicator to stop
        'Miscellaneous Ratios', 'Carbon Number Distribution', 'Parameter'}
    start_page_search_from_row: ClassVar[int] = 40
    max_cols: ClassVar[int] = 60  # look for headers in all columns up to this
    end_after_n_empty_rows: ClassVar[int] = 100
    n_empty_cols_for_row: ClassVar[int] = 10  # treat the row as empty if first n cols are
    max_rows: ClassVar[int] = 3000  # ensure an error doesn't lead to infinite loop!

    def get_pages(self) -> Iterator[GcPageData]:
        found_start = False  # not unique to each page (set once for all)
        page_no = 1
        consecutive_empty_rows = 0
        header_row_num = 0
        phase = PagePhase.Seeking
        current_page = GcPageData(str(page_no))
        header_col_to_values: DefaultDict[str, List[str]] = defaultdict(list)
        data_row_col_pair_to_value: Dict[Tuple[int, str], Any] = {}

        # 3 stopping conditions as of 05/07/2021:
        #   1) Hit >= n empty rows consecutively (after finding start of data pages)
        #      (n configured as ClassVar with {n_empty_cols_for_row})
        #   2) Detected a head for ignored data at end e.g. derived data. This is 
        #      a backup for in case there is not a large gap from end of last data page
        #      (configured as ClassVar: {post_data_headers})
        #   3) Fallback - set a max num rows just incase of a logical error or unexpected
        #      file format to prevent possible infinite loop.
        #      (max rows configured as ClassVar: {max_rows})
        for row in range(self.start_page_search_from_row, self.max_rows):  # end cond 3: max rows
            is_empty_row = self.is_empty_row(row)
            consecutive_empty_rows = consecutive_empty_rows + 1 if is_empty_row else 0

            if consecutive_empty_rows >= self.end_after_n_empty_rows and found_start:
                self.add_data_rows(current_page, data_row_col_pair_to_value)
                if current_page.has_data():
                    yield current_page
                return  # end cond 1: n consecutive empty rows

            # phase 1
            if phase == PagePhase.Seeking:
                if not is_empty_row and self.is_start_of_new_page(row):
                    found_start = True
                    phase = PagePhase.InPageMetadata

            # phase 2
            if phase == PagePhase.InPageMetadata:
                if is_empty_row:
                    phase = PagePhase.BetweenMetadataAndHeaders

            # phase 3
            if phase == PagePhase.BetweenMetadataAndHeaders:
                if not is_empty_row:
                    phase = PagePhase.InPageHeaders
                    header_row_num = 1

            # phase 4
            if phase == PagePhase.InPageHeaders:
                if header_row_num <= 2:
                    for col, val in self.get_col_value_pairs(row):
                        # handle header row 2 slight misalignment if needed...
                        col = self.adj_col_for_off_by_ones(header_row_num, 
                                                           header_col_to_values, col, val)
                        header_col_to_values[col].append(str(val))
                        if val in self.post_data_headers:
                            return  # end cond 2: into derived data (last data page already returned)
                    header_row_num += 1  
                else:
                    phase = PagePhase.InPageData

                    # add headers to current page (all values for each col to handle multi row)
                    for col, values in sorted(header_col_to_values.items(), 
                                              key=lambda kvp: col_letter_to_index(kvp[0])):
                        header = " ".join(values)
                        current_page.headers.append(header)

                    header_row_num = 0  # reset
                    header_col_to_values = defaultdict(list)  # reset

            # phase 5
            if phase == PagePhase.InPageData:
                is_new_page = self.is_start_of_new_page(row) 
                if is_empty_row or self.is_subheading(row):
                    continue
                if is_new_page:
                    if is_new_page:
                        phase = PagePhase.InPageMetadata
                    else:
                        phase = PagePhase.Seeking

                    self.add_data_rows(current_page, data_row_col_pair_to_value)

                    # return current page
                    yield current_page
                    page_no += 1
                    current_page = GcPageData(str(page_no))  # reset
                    data_row_col_pair_to_value = {}  # reset
                    continue

                for col, val in self.get_col_value_pairs(row):
                    data_row_col_pair_to_value[(row, col)] = val

    def is_empty_row(self, row: int):
        check_cols = [col_index_to_letter(i) for i in range(0, self.n_empty_cols_for_row)]
        return not any(has_value(self.read(col, row)) for col in check_cols)

    def read(self, col: str, row: int) -> Any:
        return self._xl_reader.read_cell(ExcelAddress(f"{col}{row}"))

    def is_start_of_new_page(self, row: int) -> bool:
        col_a_cleaned = str(self.read(col=self.page_start_col, row=row)).strip()
        return col_a_cleaned in self.expected_metadata_headers

    def is_subheading(self, row: int) -> bool:
        """
        Example subheadings seen in GCMSMS file provided by Paul: 
        `330.3®217.2: Internal Standard`. All examples seen only have a value in col A.
        """        
        col_value_pairs = self.get_col_value_pairs(row)
        col_A_has_value = next(col_value_pairs)[0] == 'A'
        any_other_values = any(col_value_pairs)  # any left after taking element above?
        return col_A_has_value and not any_other_values

    def get_col_value_pairs(self, row: int) -> Iterator[Tuple[str, Any]]:
        """
        Return all none empty values in row as col, value pairs.
        """
        for col_idx in range(self.max_cols):
            col = col_index_to_letter(col_idx)
            val = self.read(col, row)
            if has_value(val):
                yield col, val

    def add_data_rows(self, page: GcPageData, 
                      data_row_col_pair_to_value: Dict[Tuple[int, str], Any]) -> None:
        cols = set([c for _, c in data_row_col_pair_to_value.keys()])
        ordered_cols = sorted(cols, key=lambda col: col_letter_to_index(col))
        ordered_rows = sorted(set([r for r, _ in data_row_col_pair_to_value.keys()]))
        for r in ordered_rows:
            data_row = []
            for c in ordered_cols:
                # need to incl empties so that data lines up with headers
                # TODO: Still some risk of misalignment e.g. if one col has no values
                #       the trouble is that the header cols don't align with the values
                #       precisely (e.g. val is merged group starting one col before header col)
                #       If we get issue above could match val cols to head cols by proximity 😩
                val = data_row_col_pair_to_value.get((r, c))
                data_row.append(val)
            page.data_rows.append(data_row)

    def adj_col_for_off_by_ones(self, header_row_num, header_col_to_values, col, val):
        """
        Row 2 headers can sometimes align visually, but use a column that is off by
        one relative to the row above. If one header row 2 and the col with a value
        is not in the row 1 cols, assign to one +/- 1 col if found.
        """
        # handle off by one alignment issues i.e. col is +/- 1 on 2nd header row
        if header_row_num == 2:
            if col not in header_col_to_values:
                this_col_idx = col_letter_to_index(col)
                col_dist_pairs = [(c, abs(this_col_idx - col_letter_to_index(c)))
                                                  for c in header_col_to_values]
                off_by_ones = list(filter(lambda p: p[1] == 1, col_dist_pairs))
                if len(off_by_ones) == 1:
                    col = off_by_ones[0][0]  # correct column to match from row1
                if len(off_by_ones) > 1:
                    gc_logger.warning(f"Row 2 header {val} in col {col} does not align "
                         f"with row 1. Multiple row 1 matches were close: {off_by_ones}, "
                         f"left unchanged as it is not clear which it should join with.")
        return col
            

def get_page_reader(xl_reader: IExcelReader, file_class: GcFileClass) -> IGcPageReader:
    if file_class == GcFileClass.One:
        return GcPageClassOneReader(xl_reader)
    raise NotImplementedError()   


def merge_pages(pages: List[GcPageData]) -> GcPageData:

    def merge_two(p1: GcPageData, p2: GcPageData) -> GcPageData:
        if p1.headers != p2.headers:
            raise ValueError(f"Expected headers to be the same, but page {p2.page_no} has different "
                            f"headers.\nPage {p2.page_no}: {p2.headers}, first page: {p1.headers}.")
        p1.page_no += f",{p2.page_no}"
        p1.data_rows = p1.data_rows + p2.data_rows
        return p1

    return reduce(merge_two, pages)


def get_page_data(xl_reader: IExcelReader, file_class: GcFileClass) -> GcPageData:
    gc_logger.info("  getting page data...")
    page_reader = get_page_reader(xl_reader, file_class)
    pages = list(page_reader.get_pages())
    if not pages:
        raise StopIteration("No page data found for file.")
    gc_logger.info(f"  found {len(pages)} pages of GC data.")
    return merge_pages(pages)



import logging


logging.basicConfig()
gc_logger = logging.getLogger("GC")
gc_logger.setLevel(logging.INFO)


def get_ordinal(n: int) -> str:
    """
    Convert an integer into its ordinal representation::

        get_ordinal(0)   => '0th'
        get_ordinal(3)   => '3rd'
        get_ordinal(11)  => '11th'
        get_ordinal(21)  => '21st'
        get_ordinal(92)  => '92nd'
        get_ordinal(112) => '112th'
        get_ordinal(213) => '213th'
    """
    n = int(n)
    suffix = ['th', 'st', 'nd', 'rd', 'th'][min(n % 10, 4)]
    if 11 <= (n % 100) <= 13:
        suffix = 'th'
    return str(n) + suffix



from typing import List

import openpyxl
from openpyxl.styles import Alignment, Font
from openpyxl.workbook.workbook import Workbook

from igi_gc_reader.reader.gc_reader import IGcSheet, GcSheet, GcSheetError
from igi_gc_reader.sample_data import IGIUserFriendlyException, SampleData
from igi_gc_reader.reader.excel_address import col_index_to_letter
from igi_gc_reader.utils import gc_logger


def write_sheets(gc_sheets: List[IGcSheet], out_path: str):

    def write_sheet(wb: Workbook, idx: int, gc: GcSheet):
        sh = wb.create_sheet(index=idx, title=gc.sheet_name)
        sample_data = SampleData(gc.page_data, igi_analysis=gc.get_igi_analysis())

        context_dict = gc.context_data.get_data_dict()
        value_row = sample_data.n_header_rows + 1
        for col_idx, (header, value) in enumerate(context_dict.items()):
            col = col_index_to_letter(col_idx)
            sh[f"{col}1"] = header
            sh[f"{col}{value_row}"] = value

        headers = list(sample_data.get_headers())
        sample_row = list(sample_data.get_sample_row())
        for col_idx, ((header_1, header_2), value) in enumerate(zip(headers, sample_row), 
                                                                start=col_idx+1):
            col = col_index_to_letter(col_idx)
            sh[f"{col}1"] = header_1
            if header_2 is None:
                sh[f"{col}2"] = value
            else:
                sh[f"{col}2"] = header_2
                sh[f"{col}3"] = value

    def write_error(wb: Workbook, idx: int, gc: GcSheetError):
        sh = wb.create_sheet(index=idx, title=f"{gc.sheet_name}_err")
        sh["A1"] = "Error reading GC data:"
        sh["A1"].font = Font(bold=True)
        sh["A2"] = str(gc.error)
        sh["A2"].alignment = Alignment(wrap_text=True, vertical="top")
        sh.row_dimensions[2].height = 200
        sh.column_dimensions["A"].width = 125

        # rmv original sheet if it failed to write
        if gc.sheet_name in wb.sheetnames:
            del wb[gc.sheet_name]

    wb = openpyxl.Workbook()    
    # starting out with 1 sheet out per sheet in - can comb later if required
    existing_sht = wb.sheetnames[0]

    for idx, gc in enumerate(gc_sheets):
        try:
            gc_logger.info(f"  writing sheet data for: {gc.sheet_name}")
            if isinstance(gc, GcSheet):
                write_sheet(wb, idx, gc)
            elif isinstance(gc, GcSheetError):
                write_error(wb, idx, gc)
        except IGIUserFriendlyException as e:  # catch any errors that don't come up until writing
            gc_err = GcSheetError(gc.sheet_name, gc.file_class, e)
            write_error(wb, idx, gc_err)

    del wb[existing_sht]
    gc_logger.info(f"\nSaving output to: {out_path}")
    wb.save(out_path)



from dataclasses import dataclass, field
from typing import Iterator, Optional, DefaultDict, List, Any, Tuple
from collections import defaultdict
from igi_file_transformation_contract import IGIUserFriendlyException

from igi_gc_reader.reader.gc_page import GcPageData
from igi_gc_reader.utils import get_ordinal, gc_logger


class ColumnAlignmentException(IGIUserFriendlyException):
    pass


@dataclass
class TargetHeaders:
    a: str
    h: str
    ca: str
    ch: str


@dataclass
class SourceColumnNames:
    prop: str = field(default="")
    prop_desc: str = field(default="")
    a: str = field(default="")
    h: str = field(default="")
    ca: str = field(default="")
    ch: str = field(default="")
    ion: Optional[str] = field(default=None)

    def get_target_headers(self, prop_val: str, igi_analysis: str = "WO-GC") -> TargetHeaders:
        return TargetHeaders(
            a=f"{prop_val}[a].{igi_analysis}<count>",
            h=f"{prop_val}[h].{igi_analysis}<count>",
            ca=f"{prop_val}[ca].{igi_analysis}<{self.get_ca_uom()}>",
            ch=f"{prop_val}[ch].{igi_analysis}<{self.get_ch_uom()}>"
        )

    def get_ca_uom(self) -> str:
        return self.clean(self.ca, rmv=["area", "(", ")"])

    def get_ch_uom(self) -> str:
        return self.clean(self.ch, rmv=["height", "hght", "(", ")"])

    def clean(self, val: str, rmv: List[str]) -> str:
        val = val.lower()
        for rm in rmv:
            val = val.replace(rm, "")
        return val.strip()


Row = List[Any]


@dataclass
class SampleData:
    """Page data restructured as a single sample row."""
    page_data: GcPageData
    igi_analysis: str
    _col_head_to_values: DefaultDict[str, Row] = field(
        default_factory=lambda: defaultdict(list))
    _page_data_cols: SourceColumnNames = field(init=False)

    def __post_init__(self):
        heads = self.page_data.headers
        for row in self.page_data.data_rows:
            if len(heads) != len(row):
                msg_suffix = ""  # rows without the issue with have `None` added, in case this 
                                 # row is one of those, look for examples with more values
                if len(row) > len(heads):
                    row_with_most_values, idx = self._get_row_with_most_values()
                    nth_row = get_ordinal(idx+1)
                    msg_suffix = (f"Example row with more values ({nth_row} GC data row): \n"
                        f"{row_with_most_values}. You may need to unmerge cells to see this "
                        f"in the source spreadsheet, then remove redundant values to fix.")

                msg =(f"{len(heads)} header columns, not equal to {len(row)} values in row. "
                    f"This could be due to merged cells that contain multiple values. \n"
                    f"Headers: {heads} \n{msg_suffix}")
                raise ColumnAlignmentException(msg)
            for head, val in zip(heads, row):
                self._col_head_to_values[head].append(val)
        self._page_data_cols = self.interpret_columns()

    def get_headers(self) -> Iterator[Tuple[str, Optional[str]]]:
        cols = self._page_data_cols
        ions = self._col_head_to_values[cols.ion] if cols.ion else [
            None for _ in self._col_head_to_values[cols.prop]]
        for prop_value, prop_desc, ion in zip(self._col_head_to_values[cols.prop],
                                              self._col_head_to_values[cols.prop_desc],
                                              ions):
            use_prop = prop_value if prop_value else prop_desc
            target_headers = cols.get_target_headers(use_prop, self.igi_analysis)
            if self._page_data_cols.a:
                yield target_headers.a, ion
            if self._page_data_cols.h:
                yield target_headers.h, ion
            if self._page_data_cols.ca:
                yield target_headers.ca, ion
            if self._page_data_cols.ch:
                yield target_headers.ch, ion

    def get_sample_row(self) -> Iterator[str]:
        cols = self._page_data_cols
        d = self._col_head_to_values
        indicators_with_data = []
        if cols.a:
            indicators_with_data.append(d[cols.a])
        if cols.h:
            indicators_with_data.append(d[cols.h])
        if cols.ca:
            indicators_with_data.append(d[cols.ca])
        if cols.ch:
            indicators_with_data.append(d[cols.ch])

        for property in zip(*indicators_with_data):
            for indicator in property:
                yield indicator

    @property
    def n_header_rows(self) -> int:
        return 2 if self._page_data_cols.ion is not None else 1

    def interpret_columns(self) -> SourceColumnNames:
        """
        Used to decide how we interpret columns in the source data, which we use for
        the property, and which are assigned to each indicator.
        """
        cols = SourceColumnNames()
        heads = self.page_data.headers
        for head in heads:
            formatted_header = head.lower()
            if formatted_header == "ion":
                cols.ion = head
            elif formatted_header.startswith("peak"):
                cols.prop = head
            elif formatted_header.startswith("compound"):
                cols.prop_desc = head
            elif formatted_header.startswith("ret") and formatted_header.endswith("time"):
                continue
            elif formatted_header == "area" or ("area" in formatted_header and 
                                                "count" in formatted_header):
                cols.a = head
            elif formatted_header == "height" or ("ght" in formatted_header and 
                                                  "count" in formatted_header):
                cols.h = head
            elif "area" in formatted_header:
                # There can be multiple e.g. std and response factor corrected. Paul asked
                # for us to prefer std where both are given.
                # i.e. use if we don't have it already or if we do and existing is for resp fact
                if not cols.ca or ('RF' in cols.ca.upper() or 'RESP' in cols.ca.upper()):
                    cols.ca = head
            elif "ght" in formatted_header:  # seen "<uom> (Hght)", "<uom> (Height)" & "Hght%"
                # as above prefer std over resp fact if both are given
                if not cols.ch or ('RF' in cols.ch.upper() or 'RESP' in cols.ch.upper()):
                    cols.ch = head
            else:
                gc_logger.warning(f"Unexpected header found in page data: {head} - failed to assign")
        return cols

    def _get_row_with_most_values(self) -> Tuple[Row, int]:
        """
        All rows will have the same number of cols (None inserted to keep alignment). We 
        want to find an example of a row with the greatest number of values that are not
        None, ideally returning the first example of this i.e. lowest index.
        """
        idx_row_cnt_tuples = [(i, len(list(filter(None, r))))  # count truthy vals
                                          for i, r in enumerate(self.page_data.data_rows)]
        # sort by n_truthy_values, then by index, 1/n to invert so we can use lowest like idx
        idx_row_cnt_tuples.sort(key=lambda x: (1/(x[1]), x[0]))
        idx_with_most_values = idx_row_cnt_tuples[0][0]
        row_with_most_values = self.page_data.data_rows[idx_with_most_values]
        return row_with_most_values, idx_with_most_values


""" 
Allow easy conversion between row/col indices and Excel style addresses.
"""
from __future__ import annotations
from dataclasses import dataclass, field
from collections import namedtuple
from typing import Dict, Tuple, ClassVar


Indices: Tuple[int,int] = namedtuple('Indices', ['row', 'col'])
col_to_index_cache: Dict[str, int] = {}
index_to_col_cache: Dict[int, str] = {}


@dataclass
class ExcelAddress():
    addr: str
    indices: Indices = field(init=False)
    ASCII_A: ClassVar[int] = 65

    def __post_init__(self):
        self.indices = self.to_indices(self.addr)

    def __str__(self) -> str:
        return self.addr

    @property
    def row_idx(self):
        return self.indices.row

    @property
    def col_idx(self):
        return self.indices.col
    
    @staticmethod
    def from_indices(indices: Indices) -> ExcelAddress:
        col = col_index_to_letter(indices.col)
        return ExcelAddress(f"{col}{str(indices.row+1)}")

    @staticmethod
    def to_indices(addr: str):
        if not addr:
            return Indices(-1, -1)
        col_letters, row_num_str = "", ""
        for char in addr.upper():
            if char.isalpha():
                col_letters += char
            elif char.isdigit(): 
                row_num_str += char
            else:
                raise ValueError(f"Failed to identify char ({char}) in cell address ({addr}) "
                                 f"as either letter or number.")

        col_idx = col_letter_to_index(col_letters)
        return Indices(row=int(row_num_str)-1, col=col_idx)


def col_index_to_letter(idx: int) -> str:
    global col_to_index_cache, index_to_col_cache
    if idx in index_to_col_cache:
        return index_to_col_cache[idx]

    mod = idx % 26  # handle e.g. col letter above Z
    prefix = ""
    if idx >= 26:
        n = int(idx / 26) - 1
        prefix = chr(n + ExcelAddress.ASCII_A)

        if n >= 26:
            raise NotImplementedError("Three letter columns not yet handled")
    
    col = prefix + chr(mod + ExcelAddress.ASCII_A)

    index_to_col_cache[idx] = col
    col_to_index_cache[col] = idx
    return col


def col_letter_to_index(col: str) -> int:
    global col_to_index_cache, index_to_col_cache
    if col in col_to_index_cache:
        return col_to_index_cache[col]

    letter_indices = [ord(l)-ExcelAddress.ASCII_A for l in col]
    idx = sum([idx if pos == 0 else (idx+1) * pow(26, pos) 
                    for pos, idx in enumerate(reversed(letter_indices))])

    index_to_col_cache[idx] = col
    col_to_index_cache[col] = idx    
    return idx


def next_col(col: str) -> str:
    """i.e. D -> C, Z -> AA etc"""
    return col_index_to_letter(col_letter_to_index(col) + 1)



"""
Transform a GC lab Excel file to a format ready for p:IGI+/Metis import. 
"""
from os import stat
from igi_file_transformation_contract import (
    IFileTransformer,
    TransformationResult,
    Status,
    SuccessStatus,
    PLEASE_SUBMIT_MSG,
    IGINoSupportedSheetsException,
)
from typing import List, Optional
from collections import defaultdict
from argparse import ArgumentParser

from igi_gc_reader.reader.classification import GcFileClass
from igi_gc_reader.reader.gc_reader import IGcSheet, get_gc_sheets
from igi_gc_reader.writer import write_sheets
from igi_gc_reader.utils import gc_logger

SUPPORTED_FILE_CLASSES = [GcFileClass.One]
AUTO = "<AutoAssign>"


class GcFileTransformer(IFileTransformer):

    @property
    def title(self) -> str:
        return "GC File Transformation"

    @property
    def user_description(self) -> str:
        return "Upload a GC lab file to get an import file ready for p:IGI+ / Metis Transform."

    @property
    def accepts_file_extensions(self) -> List[str]:
        return ['.xlsx', '.xlsm', '.xls']

    def try_transform_file(self, in_path: str, out_path: str=AUTO) -> TransformationResult:
        """
        Transform GC lab file ready for p:IGI+/Transform import
        Returns: TransformationResult with path to transformed file and status
        """
        try:
            supported_sheets = get_supported_sheets(in_path)
            if (len(supported_sheets) == 0):  # if all sheets are not supported file class...
                err_msg = f"The structure of this file is not currently supported. {PLEASE_SUBMIT_MSG}"
                no_supported_sheets_err = IGINoSupportedSheetsException(err_msg)
                status = Status(success=False, igi_exception=no_supported_sheets_err)
                gc_logger.error(status.failure_message)
                return TransformationResult(status)
            out_path = self.transform_file(in_path, out_path, supported_sheets=supported_sheets)
        
        except Exception as e:  # could have a supported file class but still hit an error....
            status = Status(success=False, igi_exception=e)
            gc_logger.error(status.stack_trace)
            result = TransformationResult(status)
            return result

        result = TransformationResult(SuccessStatus, output_filepath=out_path)
        gc_logger.info(result)
        return result

    def transform_file(self, in_path: str, out_path: str = AUTO, **kwargs) -> str:
        """
        Transform GC lab file ready for p:IGI+/Transform import
        Returns: path to transformed file
        """        
        supported_sheets = kwargs.get("supported_sheets", get_supported_sheets(in_path))
        if out_path == AUTO:
            out_path = self.get_default_output_path(in_path)
        write_sheets(supported_sheets, out_path)
        return out_path

    def pick_and_transform_file(self) -> str:
        """
        Command line option to open file picker to select input file.
        Can also be used if you are using this lib within a Python script.
        For use in a web service use `transform_file` instead.
        """
        import tkinter.filedialog
        import tkinter as tk

        root = tk.Tk()  # to allow file selection dialog
        root.withdraw()  # hide root window
        input_path = tk.filedialog.askopenfilename(
            title="Select GC lab file (Excel)",
            filetypes=[("Excel files", ".xlsx .xls .xls*")],
        )
        return self.transform_file(input_path)

    @property
    def result_disclaimer(self) -> str:
        return (
            "Please check the output. We have tried to interpret the input file and "
            "assign data to appropriate indicators and units of measure. This is challenging "
            "because the headers do not always use the same names or appear in the same "
            "columns.\n\nIn the output file headers repeat in groups - one for each indicator "
            "(e.g., area, height, concentrations from area etc). If you check the first group "
            "after the context headers (well name, depth etc) the same indicators / uoms will "
            "be used for each property after."
        )

    @property
    def result_disclaimer_image_uri(self) -> Optional[str]:
        return "https://i.ibb.co/88vvsPs/disclaimer-example.png"


def get_supported_sheets(in_path: str) -> List[IGcSheet]:
    sheets_by_file_class = defaultdict(list)

    for sheet in get_gc_sheets(in_path):
        sheets_by_file_class[sheet.file_class].append(sheet)

    supported, unsupported = [], []
    for file_class in sheets_by_file_class.keys():  # i.e. if it's in supported classes
        if file_class in SUPPORTED_FILE_CLASSES:
            supported += sheets_by_file_class[file_class]
        else:
            unsupported += sheets_by_file_class[file_class]

    if len(supported) >= 1:
        if len(unsupported) == 0:
            names = [sh.sheet_name for sh in unsupported]
            gc_logger.info(f"Some unsupported sheets found: {names}")
        return supported
    return []


if __name__ == "__main__":
    parser = ArgumentParser(description=__doc__)
    parser.add_argument(
        "-i", dest="input_file_path", type=str, required=False, default=""
    )
    parser.add_argument(
        "-t", dest="use_try_func", type=str, required=False, default="n"
    )
    args = parser.parse_args()
    gc_transformer = GcFileTransformer()
    if not args.input_file_path:
        gc_transformer.pick_and_transform_file()
    else:
        if args.use_try_func.lower()[0] != "n":
            gc_transformer.try_transform_file(args.input_file_path)
        else:
            gc_transformer.transform_file(args.input_file_path)



"""
Can't use a single lib as we need to support .xls (supported by xlrd) and 
.xlsx (supported by openpyxl). Will use as thin a wrapper as possible round 
these libraries.
"""
from dataclasses import dataclass, field
from abc import ABC, abstractmethod
from typing import Any, List, Union
import os

import xlrd
from xlrd.book import Book
from xlrd.sheet import Sheet
import openpyxl
from openpyxl.workbook.workbook import Workbook
from openpyxl.xml.constants import WORKSHEET_TYPE

from igi_gc_reader.reader.excel_address import ExcelAddress


@dataclass  # type: ignore
class IExcelReader(ABC):
    filepath: str
    sheet_name: str
    _wb: Union[Workbook, Book]
    _sheet: Union[WORKSHEET_TYPE, Sheet]

    @abstractmethod
    def read_cell(self, addr: ExcelAddress) -> Any:
        pass

    def set_sheet(self, name: str) -> Any:
        sheet_names = self.get_sheet_names()
        try:
            self._sheet = self.get_sheet_by_name(name)
            self.sheet_name = name
        except:
            raise ValueError(f"No sheet named '{name}' in current workbook (sheets: {sheet_names})")

    def has_value(self, addr: ExcelAddress) -> bool:
        """Returns true if cell is not empty or whitespace only."""
        val = self.read_cell(addr)
        return has_value(val)

    @abstractmethod
    def get_sheet_names(self) -> List[str]:
        pass

    @abstractmethod
    def get_sheet_by_index(self, idx: int) -> Union[WORKSHEET_TYPE, Sheet]:
        pass

    @abstractmethod
    def get_sheet_by_name(self, name: str) -> Union[WORKSHEET_TYPE, Sheet]:
        pass

    @abstractmethod
    def close(self):
        pass


def has_value(val: Any) -> bool:
    return val and not str(val).isspace()
    

@dataclass
class ExcelOldFormatReader(IExcelReader):
    filepath: str
    sheet_name: str = field(default="")
    _wb: Book = field(init=False)
    _sheet: Sheet = field(init=False)

    def __post_init__(self):
        self._wb = xlrd.open_workbook(self.filepath)
        if self.sheet_name:
            self._sheet = self.get_sheet_by_name(self.sheet_name)
        else:
            self._sheet = self.get_sheet_by_index(0)

    def get_sheet_names(self) -> List[str]:
        return self._wb.sheet_names() 

    def read_cell(self, addr: ExcelAddress) -> Any:
        try:
            return self._sheet.cell_value(rowx=addr.row_idx, colx=addr.col_idx)
        except IndexError:
            return None

    def get_sheet_by_index(self, idx: int) -> Sheet:
        return self._wb.sheet_by_index(idx)

    def get_sheet_by_name(self, name: str) -> Sheet:
        return self._wb.sheet_by_name(name)

    def close(self):
        self._wb.release_resources()
        del self._sheet
        del self._wb


@dataclass
class ExcelNewFormatReader(IExcelReader):
    filepath: str
    sheet_name: str = field(default="")
    _wb: Workbook = field(init=False)
    _sheet: WORKSHEET_TYPE = field(init=False)

    def __post_init__(self):
        self._wb = openpyxl.load_workbook(self.filepath)
        if self.sheet_name:
            self._sheet = self.get_sheet_by_name(self.sheet_name)
        else:
            self._sheet = self.get_sheet_by_index(0)

    def get_sheet_names(self) -> List[str]:
        return self._wb.sheetnames

    def read_cell(self, addr: ExcelAddress) -> Any:
        return self._sheet[addr.addr].value

    def get_sheet_by_index(self, idx: int) -> WORKSHEET_TYPE:
        return self._wb.worksheets[idx]

    def get_sheet_by_name(self, name: str) -> WORKSHEET_TYPE:
        return self._wb[name]

    def close(self):
        self._wb.close()
        del self._sheet
        del self._wb


def get_excel_reader(filepath: str, sheet: str = "") -> IExcelReader:
    ext = os.path.splitext(filepath)[-1].lower()
    if ext == '.xls':
        return ExcelOldFormatReader(filepath, sheet)
    elif ext in ['.xlsx', '.xlsm']:
        return ExcelNewFormatReader(filepath, sheet)
    raise NotImplementedError(f"No Excel reader setup to handle {ext} extension files for file: {filepath}.")



from dataclasses import dataclass
from logging import exception
from typing import Iterator, Tuple, Union

from igi_gc_reader.reader.gc_page import GcPageData, get_page_data
from igi_gc_reader.reader.excel_reader import get_excel_reader, IExcelReader
from igi_gc_reader.reader.context_data import ContextData, get_context_data_reader
from igi_gc_reader.reader.classification import GcFileClass, classify
from igi_gc_reader.sample_data import IGIUserFriendlyException
from igi_gc_reader.utils import gc_logger
    

@dataclass(frozen=True)
class GcSheet:
    sheet_name: str
    analysis: str
    analysis_addr: str
    file_class: GcFileClass
    context_data: ContextData
    page_data: GcPageData

    def get_igi_analysis(self):
        """Map from raw analysis text in the sheet to the IGI analysis group."""
        return get_igi_analysis(self.analysis)


def get_igi_analysis(raw_analysis: str) -> str:
    """Map from raw analysis text in the sheet to the IGI analysis group."""
    cleaned = raw_analysis.lower()
    if "msms" in cleaned or "ms-ms" in cleaned:
        return "GCMS-MS"
    if "sat" in cleaned:
        return "Sat-GCMS" if "gcms" in cleaned or "bio" in cleaned else "Sat-GC"
    if "aro" in cleaned: 
        return "Arom-GCMS" if "gcms" in cleaned or "bio" in cleaned else "Arom-GC"
    if "wo" in cleaned or 'whole oil' in cleaned:
        return "WO-GCMS" if "gcms" in cleaned else "WO-GC"
    if "py" in cleaned: 
        return "Py-GC-Extr" if "ext" in cleaned else "Py-GC"
    if "ht" in cleaned or "high temp" in cleaned:
        return "HT-GC"
    if "te" in cleaned or "thermal" in cleaned:
        return "TE-GC"  # TODO - confirm whether "EXTRACT GC" should go to this or WO-GC
    if "gasol" in cleaned: 
        return "Gasol-GC"
    if "diam" in cleaned: 
        return "Diam-GCMS"
    if "pol" in cleaned: 
        return "Pol-GCMS"
    if "mrm" in cleaned: 
        return "MRM-GCMS"
    if "irms" in cleaned: 
        return "GC-IRMS"
    default = "WO-GC"
    gc_logger.warning(f"Failed to map analysis from: {raw_analysis}. Assigned to: {default}.")
    return default


@dataclass
class GcSheetError:
    sheet_name: str
    file_class: GcFileClass
    error: Exception


IGcSheet = Union[GcSheet, GcSheetError]


def build_gc_sheet(xl_reader: IExcelReader, sheet_name: str) -> GcSheet:

    def _get_analysis_and_class(xl_reader: IExcelReader) -> Tuple[GcFileClass, str, str]:
        """Returns tuple of class, analysis and analysis address"""
        result = classify(xl_reader)
        return result.file_class, result.analysis, result.analysis_addr

    gc_logger.info(f"Attempting to read sheet {sheet_name}...")
    file_class, analysis, analysis_addr = _get_analysis_and_class(xl_reader)
    context_data = get_context_data_reader(file_class, xl_reader).get_context_data()
    context_data.set_data_dict(xl_reader)
    page_data = get_page_data(xl_reader, file_class)
    return GcSheet(sheet_name, analysis, analysis_addr, file_class, context_data, page_data)


def get_gc_sheets(wb_path: str) -> Iterator[IGcSheet]:
    xl_reader = get_excel_reader(wb_path)
    for sheet in xl_reader.get_sheet_names():
        xl_reader.set_sheet(sheet)
        gc_sheet = None
        try:
            gc_sheet = build_gc_sheet(xl_reader, sheet)
            yield gc_sheet
        # ignoring these errors allows us to process workbooks with some but not all valid sheets
        except StopIteration as e:  # ignore if not page data
            gc_logger.info(f"  skipping sheet {sheet}: {e}")
        except NotImplementedError as e:  # ignore if not supported class type (handled elsewhere)
            gc_logger.info(f"  skipping sheet {sheet}: {e}")
        except IGIUserFriendlyException as e:  # proceed to next sheet but write error info
            gc_logger.error("  error in sheet {sheet}: {e}")
            cls = gc_sheet.file_class if gc_sheet is not None else GcFileClass.Unclassified
            yield GcSheetError(sheet_name=sheet, file_class=cls, error=e)
    xl_reader.close()



import setuptools


setuptools.setup(
    name="igi_gc_reader",
    version="0.1.2",
    author="IGI",
    author_email="chris.prosser@igiltd.com",
    description="GC lab file (Excel) -> Excel table formatted sheet.",
    long_description="Create files ready for p:IGI+/Transform import from print formatted GC lab files.",
    long_description_content_type="text/markdown",
    url="https://github.com/IGILtd/data_transformation_scripts/tree/master/import_file_formats/gc",
    include_package_data=True,
    license="MIT",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Database",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Pre-processors",
        "Topic :: Text Processing",
        "Typing :: Typed",
    ],
    python_requires='>=3.7',
    packages=setuptools.find_packages(exclude=["tests", "infocompiler"]),
    install_requires=["openpyxl", "xlrd", "tqdm", "igi-file-transformation-contract"]
)



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import ClassVar, Dict, List, Optional, Tuple, Any

from igi_gc_reader.reader.excel_reader import IExcelReader, has_value
from igi_gc_reader.reader.excel_address import ExcelAddress
from igi_gc_reader.reader.classification import GcFileClass


@dataclass
class ContextData:
    spec_name: str = field(default="")
    head_cols: List[str] = field(default_factory=list)
    data_cols: List[str] = field(default_factory=list)
    first_row: Optional[int] = field(default=None)
    last_row: Optional[int] = field(default=None)
    header_format: str = field(default="")  # e.g. {header:}
    _data_dict: Dict[str, str] = field(default_factory=dict)
    # get sampling date & depth first for data

    @property
    def row_range(self) -> int:
        if not self.first_row or not self.last_row:
            return -1
        return self.last_row - self.first_row

    def set_data_dict(self, xl_reader: IExcelReader) -> None:
        if self._data_dict or isinstance(self, NoContextData):
            return
        if self.first_row is None or self.last_row is None:
            raise ValueError("Need to populate first row, last row and cols before reading data.")
        for row in range(self.first_row, self.last_row + 1):
            for head_col, data_col in zip(self.head_cols, self.data_cols):
                data_addr = ExcelAddress(f"{data_col}{row}")
                val = xl_reader.read_cell(data_addr)
                if has_value(val):
                    head_addr = ExcelAddress(f"{head_col}{row}")
                    raw_header = str(xl_reader.read_cell(head_addr)).strip()
                    if raw_header.endswith(":"):
                        self.header_format = "{header}:"
                    header = raw_header.replace(':', '').title()

                    if "depth" in header.lower():
                        val, unit = self.separate_text_and_number(val)
                        header = f"{header}<{unit}>"

                    if header.lower().startswith('lat') or header.lower().startswith('lon'):
                        val = self.convert_degrees_mins_secs_to_lat_lon(val)

                    self._data_dict[header] = val


    def get_data_dict(self) -> Dict[str, str]:
        if not self._data_dict:
            raise ValueError("Need to call set_data_dict(xl_reader) to populate internal state.")
        return self._data_dict

    @staticmethod
    def convert_degrees_mins_secs_to_lat_lon(coord: str) -> Any:
        try:
            return float(coord)
        except ValueError:
            try:
                deg_str, rest = coord.split("°")
                deg = float(deg_str.strip())
                parts = rest.split("'")
                mins = float(parts[0].strip())
                secs = float(parts[1].strip())
                return deg + (mins / 60) + (secs / 60 / 60)
            except ValueError:                
                return coord  # not a recognised degrees min, sec format

    @staticmethod
    def separate_text_and_number(val: str) -> Tuple[Optional[float], str]:
        num_str, text = "", ""
        for char in val:
            if char.isdigit() or char == ".":
                num_str += char
            elif char == " ":
                continue
            else:
                text += char
        if num_str:
            return float(num_str), text
        return None, text

    def get_any_depth(self) -> str:
        data = self.get_data_dict()
        if 'Top Depth' in data:
            return data['Top Depth']
        for head, val in data.items():
            head_fmt = str(head).lower().strip()
            if 'depth' in head_fmt:
                return val
        return ""

    def get_sample_type(self) -> str:
        for head, val in self.get_data_dict().items():
            head_fmt = str(head).lower().strip()
            if head_fmt.startswith('sam') and head_fmt.endswith('type'):
                return val
        return ""


@dataclass
class NoContextData(ContextData):
    pass
    

class IContextDataReader(ABC):
    @abstractmethod
    def get_context_data(self) -> ContextData:
        pass


@dataclass
class ContextDataSpec:
    name: str
    desc: str
    header_cols: List[str]
    data_cols: List[str]


spec_weat_v1 = ContextDataSpec(name="spec_weat_v1", desc="older Weatherford style files",
                               header_cols=["A", "AC"], data_cols=["H", "AM"])

spec_weat_v2 = ContextDataSpec(name="spec_weat_v2", desc="newer Weatherford style files",
                               header_cols=["A", "AD"], data_cols=["K", "AN"])


@dataclass
class WeatherfordStyleContextDataReader(IContextDataReader):
    _xl_reader: IExcelReader
    possible_specs: ClassVar[List[ContextDataSpec]] = [spec_weat_v1, spec_weat_v2]

    def get_context_data(self) -> ContextData:
        best_range = 0
        selected_context_data: ContextData = NoContextData()
        for spec in self.possible_specs:
            start_row, end_row = 0, 0
            for row in range(1,17):  # seen start between 2-3 and end 11-14 in this class (~20 examples)
                if not start_row:
                    if all([self._xl_reader.has_value(ExcelAddress(f"{col}{row}")) 
                            for col in spec.header_cols]):
                        start_row = row
                elif not end_row:
                    if all([not self._xl_reader.has_value(ExcelAddress(f"{col}{row}")) 
                            for col in spec.header_cols]):
                        end_row = row-1
                        break
            data = ContextData(head_cols=spec.header_cols,
                               data_cols=spec.data_cols,
                               first_row=start_row, last_row=end_row)
            if data.row_range > best_range:
                selected_context_data = data
        return selected_context_data


@dataclass
class NoContextDataReader(IContextDataReader):
    def get_context_data(self) -> ContextData:
        return NoContextData()


def get_context_data_reader(file_class: GcFileClass, xl_reader: IExcelReader) -> IContextDataReader:
    if file_class == GcFileClass.One:
        return WeatherfordStyleContextDataReader(_xl_reader = xl_reader)
    # TODO - throw NotImplementedError instead and handle at top level (to report back to client)
    return NoContextDataReader()  # no reader for other file classes yet
